#include <windows.h>
#include <stdio.h>
#include <math.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <gl/glut.h>
#include <vector>
#include <list>
#include "FinalProject.h"

#define PI 3.14159265358979323

using namespace std;

// Rendering parameters
bool lightOn = true;
bool splatOn = false;
bool pointOn = true;
bool inbetweenOn = false;

// Define Grid Properties
// the unit dimensions of the voxel grid
static int gridDimX = 75;
static int gridDimY = 75;
static int gridDimZ = 75;
static int gridDimT = 12;
//the physical dimensions of the voxel grid
static float gridWidth = 10;
static float gridHeight = 10;
static float gridDepth = 10;
// the dimensions of the voxels
static float voxWidth = gridWidth/(gridDimX - 1);
static float voxHeight = gridHeight/(gridDimY - 1);
static float voxDepth = gridDepth/(gridDimZ - 1);

// animation variables
static int callbackTime = 25;
static int timeSlice = 0;
static int numSteps = 0;
static int totalFrames = 5;
static float percentComplete = 0.85;
static float stepSize = sqrt(voxWidth*voxWidth + voxHeight*voxHeight +
                             voxDepth*voxDepth)/4;


static int numPointsOnSurface = 0;
static int numPointsDoneTrans = 0;

// Types of implicit functions
static enum Surface { sphere, bumpSphere, flower };
Surface renderSurface = sphere;  // Default

// 3D vector of floats
typedef struct {
   GLfloat x;
   GLfloat y;
   GLfloat z;
} cord;

// Define voxel element
typedef struct {
   cord surfacePoint; // the implicit function value at the center of a polygonal surface face
   cord gradientPoint; // the gradient at the center of a surface face
   bool onSurface; // indicates whether or not a surface intersects the voxel

   cord transitionNormal; // stores the normal during inbetween steps
   cord transitionPosition; // stores the position during inbetween steps
   bool translatePoint; // indicates to keep updating position of point
   bool startInside; // indicates the direction the surface is moving (in, out)

   cord geoVal[8];
   GLfloat funcVal[8]; // the implicit function values at the voxel corners
   cord gradientVal[8]; // the gradient values at the voxel corners

   vector<cord> edgeList; // list of interpolated points on voxel edges with sign change
   vector<cord> gradList; // list of gradients associated with coordinates in edgelist

   GLfloat time;
} voxelCell;

// Structure to store two ints
typedef struct {
   int min0;
   int min1;
} minTwo;

// splat control variables
static int splatSides = 10;
static float splatFill = 0.8;

voxelCell ****voxelGrid = NULL;

// window dimensions
int GW, GH;

// initial viewer position
static GLdouble modelTrans[] = {0.0, 0.0, 13.0};

// initial model angle
static GLfloat theta[] = {0.0, 0.0, 0.0};
static float thetaIncr = 5.0;

//light position
GLfloat light_pos[4] = {0.0, 11.0, 0.0, 1.0};
//light color (ambient, diffuse and specular)
GLfloat light_amb[4] = {0.6, 0.6, 0.6, 1.0};
GLfloat light_diff[4] = {0.6, 0.6, 0.6, 1.0};
GLfloat light_spec[4] = {0.8, 0.8, 0.8, 1.0};

//set up some materials
typedef struct materialStruct {
  GLfloat ambient[4];
  GLfloat diffuse[4];
  GLfloat specular[4];
  GLfloat shininess[1];
} materialStruct;

materialStruct GreenShiny = {
  {0.0, 0.3, 0.0, 1.0},
  {0.0, 0.9, 0.0, 1.0},
  {0.2, 1.0, 0.2, 1.0},
  {8.0}
};

// Allocate memory for voxel grid structure
void newGridMemory(int dimX, int dimY, int dimZ, int dimT) {
   voxelGrid = (voxelCell ****)malloc(dimX*sizeof(vector<voxelCell> ***));

   for(int x = 0; x < dimX; x++) {
      voxelGrid[x] = (voxelCell ***)malloc(dimY*sizeof(voxelCell **));

      for(int y = 0; y < dimY; y++) {
         voxelGrid[x][y] = (voxelCell **)malloc(dimZ*sizeof(voxelCell *));
         
         for(int z = 0; z < dimZ; z++) {
            voxelGrid[x][y][z] = (voxelCell *)malloc(dimT*sizeof(voxelCell));
         }
      }
   }
}

//---------------------------------------------------------
//  Utility functions

void normalize(cord *v) {
   // normalize v[] and return the result in v[]
   // from OpenGL Programming Guide, p. 58
   GLfloat d = sqrt(v->x*v->x + v->y*v->y + v->z*v->z);
   if (d == 0.0) {
      printf("zero length vector");
      return;
   }
   v->x = v->x/d;
   v->y = v->y/d;
   v->z = v->z/d;
}

cord normCrossProd(cord v1, cord v2) {
   // cross v1[] and v2[] and return the result in out[]
   // from OpenGL Programming Guide, p. 58
   cord out;
   out.x = v1.y*v2.z - v1.z*v2.y;
   out.y = v1.z*v2.x - v1.x*v2.z;
   out.z = v1.x*v2.y - v1.y*v2.x;
   normalize(&out);
   return out;
}

float normDotProd(cord v1, cord v2) {
   // cross v1[] and v2[] and return the result in out[]
   // from OpenGL Programming Guide, p. 58
   normalize(&v1);
   normalize(&v2);

   return v1.x*v2.x + v1.y*v2.y + v1.z*v2.z;
}

//---------------------------------------------------------
//  Grid setup functions

// sets up the geometry of a voxel
void setGeometry(int i, int j, int k, cord *geo) {
   geo[0].x = i * voxWidth;
   geo[0].y = j * voxHeight;
   geo[0].z = k * voxDepth;

   geo[1].x = i * voxWidth;
   geo[1].y = (j + 1) * voxHeight;
   geo[1].z = k * voxDepth;

   geo[2].x = i * voxWidth;
   geo[2].y = (j + 1) * voxHeight;
   geo[2].z = (k + 1) * voxDepth;

   geo[3].x = i * voxWidth;
   geo[3].y = j * voxHeight;
   geo[3].z = (k + 1) * voxDepth;

   geo[4].x = (i + 1) * voxWidth;
   geo[4].y = j * voxHeight;
   geo[4].z = k * voxDepth;

   geo[5].x = (i + 1) * voxWidth;
   geo[5].y = (j + 1) * voxHeight;
   geo[5].z = k * voxDepth;

   geo[6].x = (i + 1) * voxWidth;
   geo[6].y = (j + 1) * voxHeight;
   geo[6].z = (k + 1) * voxDepth;

   geo[7].x = (i + 1) * voxWidth;
   geo[7].y = j * voxHeight;
   geo[7].z = (k + 1) * voxDepth;
}

// Assign values to voxel 
void setVoxelValues(float ***funcVals, int t) {
   for (int i = 0; i < gridDimX; i++) {
      for (int j = 0; j < gridDimY; j++) {
         for (int k = 0; k < gridDimZ; k++) {
            setGeometry(i, j, k, voxelGrid[i][j][k][t].geoVal);
            voxelGrid[i][j][k][t].time = t;
            voxelGrid[i][j][k][t].onSurface = false;

            voxelGrid[i][j][k][t].funcVal[0] = funcVals[i][j][k];
            voxelGrid[i][j][k][t].funcVal[1] = funcVals[i][j+1][k];
            voxelGrid[i][j][k][t].funcVal[2] = funcVals[i][j+1][k+1];
            voxelGrid[i][j][k][t].funcVal[3] = funcVals[i][j][k+1];
            voxelGrid[i][j][k][t].funcVal[4] = funcVals[i+1][j][k];
            voxelGrid[i][j][k][t].funcVal[5] = funcVals[i+1][j+1][k];
            voxelGrid[i][j][k][t].funcVal[6] = funcVals[i+1][j+1][k+1];
            voxelGrid[i][j][k][t].funcVal[7] = funcVals[i+1][j][k+1];
         }
      }
   }
}

// calculate the gradient at a corner
cord getGradient(float xPlusOne, float xMinOne, float yPlusOne, float yMinOne,
                 float zPlusOne, float zMinOne, float deltaX, float deltaY, float deltaZ) {
   cord gradient;

   gradient.x = (xPlusOne - xMinOne)/deltaX;
   gradient.y = (yPlusOne - yMinOne)/deltaY;
   gradient.z = (zPlusOne - zMinOne)/deltaZ;

   return gradient;
}

// get the gradients for voxel corners
void setGradientValues(int t) {
   for (int i = 1; i < gridDimX - 1; i++) {
      for (int j = 1; j < gridDimY - 1; j++) {
         for (int k = 1; k < gridDimZ - 1; k++) {
            for (int index = 0; index < 8; index++) {
               voxelGrid[i][j][k][t].gradientVal[index] = 
                  getGradient(voxelGrid[i-1][j][k][t].funcVal[index],
                              voxelGrid[i+1][j][k][t].funcVal[index],
                              voxelGrid[i][j-1][k][t].funcVal[index],
                              voxelGrid[i][j+1][k][t].funcVal[index],
                              voxelGrid[i][j][k-1][t].funcVal[index],
                              voxelGrid[i][j][k+1][t].funcVal[index],
                              voxWidth, voxHeight, voxDepth);
            }
         }
      }
   }
}

// populate the grid with implicit function values
void setupGrid() {
   // Allocate pointer array
   int pointCountX = gridDimX + 1;
   int pointCountY = gridDimY + 1;
   int pointCountZ = gridDimZ + 1;

   // Allocate point memory
   float ***funcVals = (float ***)malloc(pointCountX*sizeof(float **));
   
   for (int x = 0; x < pointCountX; x++) {
      funcVals[x] = (float **)malloc(pointCountY*sizeof(float *));

      for (int y = 0; y < pointCountY; y++) {
         funcVals[x][y] = (float *)malloc(pointCountZ*sizeof(float));
      }
   }

   // For each timestep
   for (int t = 0; t < gridDimT; t++) {
      // Compute function values
      for (int i = 0; i < pointCountX; i++) {
         for (int j = 0; j < pointCountY; j++) {
            for (int k = 0; k < pointCountZ; k++) {
               funcVals[i][j][k] = implicitFunction(i, j, k, t);
            }
         }
      }
      // Assign values to voxel
      setVoxelValues(funcVals, t);
      setGradientValues(t);
   }

   // Free point memory
   for (int x = 0; x < pointCountX; x++) {
      for (int y = 0; y < pointCountY; y++) {
         free(funcVals[x][y]);
      }
      free(funcVals[x]);
   }
}

// determine whether an edge has a sign change
bool hasSignChange(float p1, float p2) {
   return !(p1 < 0 && p2 < 0) &&
          !(p1 >= 0 && p2 >= 0);
}

// determines if a point is inside the implicit surface of slice t
bool pointInsideSurface(int t, cord pt) {
   int x = pt.x/voxWidth;
   int y = pt.y/voxHeight;
   int z = pt.z/voxDepth;

   voxelCell *vox = &voxelGrid[x][y][z][t];

   bool hasNeg = false;
   int index = 0;
   while (!hasNeg && index < 8) {
      hasNeg = (vox->funcVal[index] < 0);
      index++;
   }

   return hasNeg;
}

// calculate the gradient at a point
cord getGradient(int t, cord pt) {
   int x = pt.x/voxWidth;
   int y = pt.y/voxHeight;
   int z = pt.z/voxDepth;

   return voxelGrid[x][y][z][t].gradientPoint;
}

// interpolate between two coordinates
cord interpolateCord(cord p1, cord p2, float t) {
   cord newCord;

   newCord.x = p1.x + t * (p2.x - p1.x);
   newCord.y = p1.y + t * (p2.y - p1.y);
   newCord.z = p1.z + t * (p2.z - p1.z);

   return newCord;
}

// determine whether the implicit surface intersects a voxel edge
void addEdgePoints(int startIndex, int endIndex, float *pt, cord *geo, vector<cord> *edgeList,
                   cord *grad, vector<cord> *gradList) {
   // If the sign changes between the corner points
   float startPt = pt[startIndex];
   float endPt = pt[endIndex];

   if (hasSignChange(startPt, endPt)) {
      float t = (0.0 - startPt)/(endPt - startPt);

      // Add interpolated cordinate to list
      edgeList->push_back(interpolateCord(geo[startIndex], geo[endIndex], t));
      gradList->push_back(interpolateCord(grad[startIndex], grad[endIndex], t));
   }
}

// calculate the Euclidean distance between two points
float getCordDistance(cord p1, cord p2) {
   float x = p1.x - p2.x;
   float y = p1.y - p2.y;
   float z = p1.z - p2.z;

   return sqrt(x*x + y*y + z*z);
}

// build a triangle by determining the two closest vertices to the initial vertex
minTwo twoClosest(cord *pt, vector<cord> *edgeList) {
   float min[2] = {-1, -1};
   minTwo minCord;

   for(int i = 0; i < (int)edgeList->size(); i++) {
      float dist = getCordDistance(*pt, (*edgeList)[i]);
      if (min[0] == -1) {
         min[0] = dist;
         minCord.min0 = i;
      }
      else if (min[1] == -1) {
         min[1] = dist;
         minCord.min1 = i;
      }
      else if (dist < min[0]) {
         min[1] = min[0];
         minCord.min1 = minCord.min0;

         min[0] = dist;
         minCord.min0 = i;
      }
      else if (dist < min[1]) {
         min[1] = dist;
         minCord.min1 = i;
      }
   }
   return minCord;
}

// find center of a triangle
cord centerOfTriangle(vector<cord> *list) {
   cord end = list->back();
   return interpolateCord(end, interpolateCord((*list)[0], (*list)[1], 0.5), 0.5);
}

// find the center of a quadrilateral
cord centerOfQuad(vector<cord> *list) {
   vector<cord> *triangle = new vector<cord>;

   cord removePt = list->back();
   list->pop_back();

   minTwo ptZeroMin = twoClosest(&removePt, list);
   triangle->push_back((*list)[ptZeroMin.min0]);
   triangle->push_back((*list)[ptZeroMin.min1]);
   triangle->push_back(removePt);

   cord ret = interpolateCord(centerOfTriangle(triangle), centerOfTriangle(list), 0.5);
   delete triangle;

   return ret;
}

// find the center of a pentagon
cord centerOfPenta(vector<cord> *list) {
   vector<cord> *triangle = new vector<cord>;

   cord removePt = list->back();
   list->pop_back();

   minTwo ptZeroMin = twoClosest(&removePt, list);
   triangle->push_back((*list)[ptZeroMin.min0]);
   triangle->push_back((*list)[ptZeroMin.min1]);
   triangle->push_back(removePt);

   cord ret = interpolateCord(centerOfTriangle(triangle), centerOfQuad(list), 0.5);
   delete triangle;

   return ret;
}

// find the center of a hexagon
cord centerOfHex(vector<cord> *list) {
   vector<cord> *triangle = new vector<cord>;

   cord removePt = list->back();
   list->pop_back();

   minTwo ptZeroMin = twoClosest(&removePt, list);
   triangle->push_back((*list)[ptZeroMin.min0]);
   triangle->push_back((*list)[ptZeroMin.min1]);
   triangle->push_back(removePt);

   cord ret = interpolateCord(centerOfTriangle(triangle), centerOfPenta(list), 0.5);
   delete triangle;

   return ret;
}

// find gradientPoint and surfacePoint for a voxel
void interpolateGridPoints() {
   for (int t = 0; t < gridDimT; t++) {
      for (int i = 0; i < gridDimX; i++) {
         for (int j = 0; j < gridDimY; j++) {
            for (int k = 0; k < gridDimZ; k++) {
               float *pt = voxelGrid[i][j][k][t].funcVal;
               cord *geo = voxelGrid[i][j][k][t].geoVal;
               cord *grad = voxelGrid[i][j][k][t].gradientVal;
               
               vector<cord> *edgeList = &voxelGrid[i][j][k][t].edgeList;
               vector<cord> *gradList = &voxelGrid[i][j][k][t].gradList;

               edgeList->clear();
               gradList->clear();
               
               // Get surface intersection coordinates
               // on voxel edges
               addEdgePoints(0, 1, pt, geo, edgeList, grad, gradList);
               addEdgePoints(0, 3, pt, geo, edgeList, grad, gradList);
               addEdgePoints(0, 4, pt, geo, edgeList, grad, gradList);
               addEdgePoints(1, 2, pt, geo, edgeList, grad, gradList);
               addEdgePoints(1, 5, pt, geo, edgeList, grad, gradList);
               addEdgePoints(2, 3, pt, geo, edgeList, grad, gradList);
               addEdgePoints(2, 6, pt, geo, edgeList, grad, gradList);
               addEdgePoints(3, 7, pt, geo, edgeList, grad, gradList);
               addEdgePoints(4, 5, pt, geo, edgeList, grad, gradList);
               addEdgePoints(4, 7, pt, geo, edgeList, grad, gradList);
               addEdgePoints(5, 6, pt, geo, edgeList, grad, gradList);
               addEdgePoints(6, 7, pt, geo, edgeList, grad, gradList);

               voxelCell *voxel = &voxelGrid[i][j][k][t];
               
               cord center, centerGrad;

               int listSize = edgeList->size();

			   // find the center of the polygon by iteratively splitting it into triangles
               if (listSize == 3) {
                  center = centerOfTriangle(edgeList);
                  centerGrad = centerOfTriangle(gradList);
               }
               else if (listSize == 4) {
                  center = centerOfQuad(edgeList);
                  centerGrad = centerOfQuad(gradList);
               }
               else if (listSize == 5) {
                  center = centerOfPenta(edgeList);
                  centerGrad = centerOfPenta(gradList);
               }
               else if (listSize == 6) {
                  center = centerOfHex(edgeList);
                  centerGrad = centerOfHex(gradList);
               }
               else if (listSize != 0) {
                  printf("ERROR: Number of points on edges not 3 or 4\n");
                  printf("Edge count: %i\n", (*edgeList).size());
			   }

               if (listSize >= 3 && listSize <= 6) {
                  voxel->onSurface = true;

                  voxel->surfacePoint.x = center.x;
                  voxel->surfacePoint.y = center.y;
                  voxel->surfacePoint.z = center.z;

                  voxel->gradientPoint.x = centerGrad.x;
                  voxel->gradientPoint.y = centerGrad.y;
                  voxel->gradientPoint.z = centerGrad.z;
               }
			   // the surface doesn't intersect the voxel
			   // calculate gradient value by averaging corners
               else {
                  float x, y, z;
                  x = y = z = 0.0;

                  for (int i = 0; i < 8; i++) {
                     x += grad[i].x;
                     y += grad[i].y;
                     z += grad[i].z;
                  }
                  voxel->gradientPoint.x = x/8;
                  voxel->gradientPoint.y = y/8;
                  voxel->gradientPoint.z = z/8;
               }
            }
         }
      }
   }
}

//---------------------------------------------------------
//   Set up the view

void setUpView() {
   // this code initializes the viewing transform
   glLoadIdentity();

   // moves viewer along coordinate axes
   gluLookAt(0.0, 0.0, 15.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);

   // move the view back some relative to viewer[] position
   glTranslatef(0.0f,0.0f, -8.0f);

   // rotates view
   glRotatef(0, 1.0, 0.0, 0.0);
   glRotatef(0, 0.0, 1.0, 0.0);
   glRotatef(0, 0.0, 0.0, 1.0);

   return;
}

//----------------------------------------------------------
//  Setup model transform

void setUpModelTransform() {
   // moves model along coordinate axes
   glTranslatef(modelTrans[0], modelTrans[1], modelTrans[2]);

   // rotates model
   glRotatef(theta[0], 1.0, 0.0, 0.0);
   glRotatef(theta[1], 0.0, 1.0, 0.0);
   glRotatef(theta[2], 0.0, 0.0, 1.0);
}

// Setup a specific material
void materials(materialStruct materials) {
  glMaterialfv(GL_FRONT, GL_AMBIENT, materials.ambient);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, materials.diffuse);
  glMaterialfv(GL_FRONT, GL_SPECULAR, materials.specular);
  glMaterialfv(GL_FRONT, GL_SHININESS, materials.shininess);
}

// Setup lighting
void init_lighting() {
  //turn on light0
  glEnable(GL_LIGHT0);
  //set up the diffuse, ambient and specular components for the light
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diff);
  glLightfv(GL_LIGHT0, GL_AMBIENT, light_amb);
  glLightfv(GL_LIGHT0, GL_SPECULAR, light_spec);
  //specify our lighting model as 1 normal per face
  glShadeModel(GL_SMOOTH);
}

// Setup light position
void pos_light() {
  //set the light's position
  glMatrixMode(GL_MODELVIEW);
  glLightfv(GL_LIGHT0, GL_POSITION, light_pos);
}

//------------------------------------------------------------------------------
//  Implicit Function Definitions

// get the implicit value of the sphere at (x,y,z,t)
float sphereFunc(float radius, int x, int y, int z) {
   float X = (float)x*voxWidth - (float)gridWidth/2.0;
   float Y = (float)y*voxHeight - (float)gridHeight/2.0;
   float Z = (float)z*voxDepth - (float)gridDepth/2.0;

	return X*X + Y*Y + Z*Z - radius*radius;
}

// get the implicit value of the bump sphere at (x,y,z,t)
float bumpSphereFunc(float radius, int x, int y, int z, float a, float b) {
   float X = (float)x*voxWidth - (float)gridWidth/2.0;
   float Y = (float)y*voxHeight - (float)gridHeight/2.0;
   float Z = (float)z*voxDepth - (float)gridDepth/2.0;

   return X*X + Y*Y + Z*Z + a*cos(b*x) + a*cos(b*y) + a*cos(b*z) -
          radius*radius;
}

// get the implicit value of the flower function at (x,y,z,t)
float flowerFunc(int x, int y, int z, int a, int b) {
   float X = (float)x * voxWidth - (float)gridWidth/2.0;
   float Y = (float)y * voxHeight - (float)gridHeight/2.0;
   float Z = (float)z * voxDepth - (float)gridDepth/2.0;

   double radius = sqrt(X*X + Y*Y + Z*Z);
   float theta = atan2(X, Z);
   float phi = atan2(X*X + Z*Z, Y);

   return gridWidth/2.0 * sin(a*theta) * sin(b*phi) - radius;
}

// get the value of the implicit function at (x,y,z,t)
float implicitFunction(int x, int y, int z, int t) {
   float implicitReturn;

   float deltaT = 0.0;
   if (gridDimT > 1) {
      deltaT = (float)t/(float)(gridDimT-1);
   }

   if (renderSurface == bumpSphere) {
      float radius = 4.0 + (1.0 - 4.0)*deltaT;
      float a = 0.6 + (0.0 - 0.6)*deltaT;
      float b = 7.4;
      implicitReturn = bumpSphereFunc(radius, x, y, z, a, b);
   }
   else if (renderSurface == flower) {
      int a = 1 + (5-1)*deltaT;
      implicitReturn = flowerFunc(x, y, z, a, 4);
   }
   else {
      float radius = 1.0 + (4.0 - 1.0)*deltaT;
      implicitReturn = sphereFunc(radius, x, y, z);
   }

   return implicitReturn;
}

//------------------------------------------------------------------------------
// Draw Functions

// draw splats
void drawSplat(GLfloat radius, cord center, cord normal) {
   glPushMatrix();
      normalize(&normal);

      cord splatNormal;
      splatNormal.x = 0.0;
      splatNormal.y = 1.0;
      splatNormal.z = 0.0;

      cord rotAxis = normCrossProd(normal, splatNormal);
      double dot = (double)normDotProd(normal, splatNormal);
      float angle = acos(dot);

      glTranslatef(center.x, center.y, center.z);
      glRotatef(angle*180.0/PI, -rotAxis.x, -rotAxis.y, -rotAxis.z);

      glBegin(GL_POLYGON);

      for (int i = 0; i < splatSides; i++) {
         float angle = -(float)(2 * PI* i) / splatSides;
         glNormal3f(0.0, 1.0, 0.0);
         glVertex3f(radius * cos(angle), 0.0, radius * sin(angle));
      }

      glEnd();
   glPopMatrix();
}

// draw the implicit surface
void drawImplicitFunction() {
   materials(GreenShiny);

   int t = timeSlice;
	// iterate through the voxel grid
	for(int i = 0; i < gridDimX; i++) {
      for(int j = 0; j < gridDimY; j++) {
         for(int k = 0; k < gridDimZ; k++) {
            voxelCell *voxel = &voxelGrid[i][j][k][t];
			// if the surface intersects the voxel
            if (voxel->onSurface) {
			   // the surface at the initial timestep, before stepping along gradient
               if (numSteps == 0) {
                  voxel->transitionNormal = voxel->gradientPoint;
                  normalize(&(voxel->transitionNormal));
                  voxel->transitionPosition = voxel->surfacePoint;
                  voxel->translatePoint = true;
                  voxel->startInside = pointInsideSurface(t+1, voxel->transitionPosition);
                  
                  numPointsOnSurface++;
               }

               cord* normal = &voxel->transitionNormal;
               cord* point = &voxel->transitionPosition;
			   // check if point has reached the surface
               if (voxel->translatePoint) {
                  bool pointInside = pointInsideSurface(t + 1, *point);
                  if (!voxel->startInside &&  pointInside ||
                       voxel->startInside && !pointInside) {
                     voxel->translatePoint = false;
                     numPointsDoneTrans++;
                  }
               }
			   // transitioning between time steps
               if (t < (gridDimT - 1) && inbetweenOn && voxel->translatePoint) {

                  normal = &getGradient(t + 1, *point);
                  normalize(normal);

                  int direction;
                  if (!voxel->startInside) {
                     direction = 1;
                  }
                  else {
                     direction = -1;
                  }
				  // step along gradient
                  point->x += normal->x * stepSize * direction;
                  point->y += normal->y * stepSize * direction;
                  point->z += normal->z * stepSize * direction;
               }

               // Render points
               if (pointOn) {
                  glColor3f(1.0, 0.0, 0.0);
                  glBegin(GL_POINTS);
                     glNormal3f(normal->x, normal->y, normal->z);
                     glVertex3f(point->x, point->y, point->z);
                  glEnd();
               }
               // Render splats
               if (splatOn) {
                  glColor3f(0.0, 1.0, 1.0);
                  float radius = getCordDistance(voxel->edgeList[0], voxel->surfacePoint);
                  drawSplat(radius*splatFill, *point, *normal);
               }
            }
         }
      }
   }
   numSteps++;

   if (!inbetweenOn && numSteps > totalFrames || inbetweenOn &&
       (float)numPointsDoneTrans/(float)numPointsOnSurface > percentComplete) {
      numSteps = 0;
      numPointsOnSurface = 0;
      numPointsDoneTrans = 0;
      timeSlice++;
   }
   if (!inbetweenOn && timeSlice >= gridDimT ||
       inbetweenOn && timeSlice >= gridDimT - 1) {
      timeSlice = 0;
   }
}

// setup the voxel grid geometry
void drawBox() {
   glDisable(GL_LIGHTING);
      
	glBegin(GL_LINES);
      
	  glColor3f(1.0, 1.0, 1.0);
   
      glVertex3f(0.0, 0.0, 0.0);
      glVertex3f(0.0, gridHeight, 0.0);

      glVertex3f(0.0, gridHeight, 0.0);
      glVertex3f(gridWidth, gridHeight, 0.0);

      glVertex3f(gridWidth, gridHeight, 0.0);
      glVertex3f(gridWidth, 0.0, 0.0);

      glVertex3f(gridWidth, 0.0, 0.0);
      glVertex3f(0.0, 0.0, 0.0);


      glVertex3f(0.0, 0.0, gridDepth);
      glVertex3f(0.0, gridHeight, gridDepth);

      glVertex3f(0.0, gridHeight, gridDepth);
      glVertex3f(gridWidth, gridHeight, gridDepth);

      glVertex3f(gridWidth, gridHeight, gridDepth);
      glVertex3f(gridWidth, 0.0, gridDepth);

      glVertex3f(gridWidth, 0.0, gridDepth);
      glVertex3f(0.0, 0.0, gridDepth);


      glVertex3f(0.0, 0.0, 0.0);
      glVertex3f(0.0, 0.0, gridDepth);
      
      glVertex3f(gridWidth, 0.0, 0.0);
      glVertex3f(gridWidth, 0.0, gridDepth);

      glVertex3f(0.0, gridHeight, 0.0);
      glVertex3f(0.0, gridHeight, gridDepth);

      glVertex3f(gridWidth, gridHeight, 0.0);
      glVertex3f(gridWidth, gridHeight, gridDepth);

	glEnd( );
	
   if (lightOn)
	   glEnable(GL_LIGHTING);
}

void display (void) {
 
   // this code executes whenever the window is redrawn (when opened,
   //   moved, resized, etc.
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   // set the viewing transform
   setUpView();

   // start drawing objects
   setUpModelTransform();
   
   glPushMatrix();
      glTranslatef(-(float)gridWidth/2.0, -(float)gridHeight/2.0, -(float)gridDepth/2.0);
      drawImplicitFunction();
      drawBox();
   glPopMatrix();
   glutSwapBuffers();

}

//-----------------------------------------------------------
//  Callback functions

void reshapeCallback(int w, int h) {
   GW = w;
   GH = h;

   glViewport(0, 0 ,w,h);

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   
   if (w < h) {
      glFrustum(-2.0, 2.0, -2.0*(GLfloat) h / (GLfloat) w,
                2.0*(GLfloat) h / (GLfloat) w, 2.0, 20.0);
   }
   else {
      glFrustum(-2.0, 2.0, -2.0*(GLfloat) w / (GLfloat) h,
                2.0*(GLfloat) w / (GLfloat) h, 2.0, 20.0);
   }

   glMatrixMode(GL_MODELVIEW);
}

void mouseCallback(int button, int state, int x, int y) {
   // rotate camera
   GLint axis = 3;
   if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) axis = 0;
   if (button == GLUT_MIDDLE_BUTTON && state == GLUT_DOWN) axis = 1;
   if (button == GLUT_RIGHT_BUTTON && state == GLUT_DOWN) axis = 2;
   if (axis < 3) {  // button ups won't change axis value from 3
      theta[axis] += thetaIncr;
      if (theta[axis] > 360.0) theta[axis] -= 360.0;
      display();
   }
}   

void keyCallback(unsigned char key, int x, int y) {

   switch( key ) {
      // model translations
      case 'x':   modelTrans[0] -= 1.0;   break;
      case 'X':   modelTrans[0] += 1.0;   break;
      case 'y':   modelTrans[1] -= 1.0;   break;
      case 'Y':   modelTrans[1] += 1.0;   break;
      case 'z':   modelTrans[2] -= 1.0;   break;
      case 'Z':   modelTrans[2] += 1.0;   break;
      case 'r':   theta[0] = theta[1] = theta[2] = 0.0;   break;
	  // model rotations
      case '-':   thetaIncr = -thetaIncr; break;
      case '+':
         if (thetaIncr < 0) thetaIncr = thetaIncr - 1.0;
         else               thetaIncr = thetaIncr + 1.0;
         break;
		 // turn on/off illumination
         case 'l': case 'L':
         lightOn = !lightOn;
         if (lightOn) {
            glEnable(GL_LIGHTING);
            printf("Lighting ON\n");
         }
         else {
            glDisable(GL_LIGHTING);
            printf("Lighting OFF\n");
         }
         break;
	  // show splats
      case 's': case 'S':  splatOn = !splatOn;  break;
	  // show points
      case 'p': case 'P':  pointOn = !pointOn;  break;
	  // 
      case 'i': case 'I':  inbetweenOn = !inbetweenOn;  break;
      // draw the 'flower' function
      case 'f': case 'F':
         renderSurface = flower;
         setupGrid();
         interpolateGridPoints();
         numSteps = 0;
         numPointsOnSurface = 0;
         numPointsDoneTrans = 0;
         timeSlice = 0;
         break;
	  // draw the 'bump sphere'
      case 'b': case 'B':
         renderSurface = bumpSphere;
         setupGrid();
         interpolateGridPoints();
         numSteps = 0;
         numPointsOnSurface = 0;
         numPointsDoneTrans = 0;
         timeSlice = 0;
         break;
	  // draw the normal sphere
      case 'e': case 'E':
         renderSurface = sphere;
         setupGrid();
         interpolateGridPoints();
         numSteps = 0;
         numPointsOnSurface = 0;
         numPointsDoneTrans = 0;
         timeSlice = 0;
         break;
	  // exit the program
      case 'q': case 'Q' : exit( EXIT_SUCCESS );   break;
   }
   glutPostRedisplay();
}

void timeStep(int step) {
   display();
   glutTimerFunc(callbackTime, timeStep, 0);
}

//------------------------------------------------------------------------------

int main(int argc, char* argv[]) {
   newGridMemory(gridDimX, gridDimY, gridDimZ, gridDimT);
   setupGrid();
   interpolateGridPoints();

    //initialize globals
   GW = 700;
   GH = 700;
   
   //set up glut window
   glutInit(&argc, argv);
   glutInitWindowSize(GW, GH);
   glutInitWindowPosition(100, 100);
   glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
   glutCreateWindow("Implicit Surface");
   glClearColor(0.0, 0.0, 0.0, 1.0);

   //set up glut callbacks
   glutDisplayFunc(display);
   glutReshapeFunc(reshapeCallback);
   glutKeyboardFunc(keyCallback);
   glutMouseFunc(mouseCallback);
   glutTimerFunc(callbackTime, timeStep, 0);
  
   //enable GL features we want
   glEnable(GL_DEPTH_TEST);
   
   if (lightOn)
      glEnable(GL_LIGHTING);

   glCullFace(GL_BACK);
   glEnable(GL_CULL_FACE);

   init_lighting();

   glutMainLoop();
   return 0;
}